# -*- coding: utf-8 -*-
from imio.fpaudit.storage import store_config
from imio.helpers.security import get_zope_root
from imio.helpers.security import set_site_from_package_config
from plone import api

import os
import transaction


def zope_ready(event):
    """Not going here in test"""
    zope_app = get_zope_root()
    site = set_site_from_package_config("imio.fpaudit", zope_app=zope_app)
    change = False
    if site:
        if os.getenv("INSTANCE_HOME", "").endswith("/instance1") or os.getenv("INSTANCE_HOME", "").endswith(
            "/instance"
        ):
            with api.env.adopt_user("admin"):
                log_entries = api.portal.get_registry_record("imio.fpaudit.settings.log_entries", default=[])
                if log_entries:
                    store_config(log_entries)
                    change = True
    if change:
        transaction.commit()



# -*- coding: utf-8 -*-
"""Init and utils."""
from zope.i18nmessageid import MessageFactory

import os


_ = MessageFactory("imio.fpaudit")

if os.environ.get("ZOPE_HOME", ""):
    BLDT_DIR = "/".join(os.getenv("INSTANCE_HOME", "").split("/")[:-2])
else:  # test env
    BLDT_DIR = os.getenv("PWD", "")

LOG_DIR = BLDT_DIR
if os.path.exists(os.path.join(BLDT_DIR, "var/log")):
    LOG_DIR = os.path.join(BLDT_DIR, "var/log")



from imio.fpaudit import LOG_DIR
from imio.fpaudit.interfaces import ILogsStorage
from imio.fpaudit.logger import FPAuditLogInfo
from zope.component import getUtility
from zope.interface import implementer

import os


@implementer(ILogsStorage)
class LogsStorageUtility(object):
    """Utility to store logs instances"""

    def __init__(self):
        self.storage = {}

    def add(self, key, value):
        """Add a log entry in storage."""
        self.storage[key] = value

    def set(self, dic):
        """Set storage to dic"""
        self.storage = dic

    def get(self, key, default=None):
        """Get a key from storage"""
        return self.storage.get(key, default)

    def remove(self, key):
        """Remove key from storage"""
        if key in self.storage:
            del self.storage[key]


def store_config(logs_config):
    """Store logs configuration in utility."""
    dic = {}
    for entry in logs_config:
        log_i = FPAuditLogInfo(
            {"audit-log": os.path.join(LOG_DIR, entry["audit_log"])},
            entry["log_id"],
            logformat=entry["log_format"],
        )
        log_i.handler.formatter.datefmt = "%Y-%m-%d %H:%M:%S"
        dic[entry["log_id"]] = log_i
    storage = getUtility(ILogsStorage)
    storage.set(dic)



from collective.fingerpointing.config import LOG_FORMAT
from collective.fingerpointing.logger import LogInfo

import logging


class FPAuditLogInfo(LogInfo):
    def __init__(self, config, log_id, logformat=LOG_FORMAT):  # noqa
        self.logger = logging.getLogger("fpa_{}".format(log_id))
        self.logfile = None
        self.handler = None
        self.configure(config, logformat=logformat)



from collective.documentgenerator.helper import DocumentGenerationHelperView
from imio.fpaudit import LOG_DIR
from imio.fpaudit.utils import get_all_lines_of
from imio.fpaudit.utils import get_lines_info
from imio.fpaudit.utils import get_log_option
from imio.fpaudit.utils import get_logrotate_filenames


class FPAuditHelperView(DocumentGenerationHelperView):
    """
    Helper methods used for document generation
    """

    def get_log_lines(self):
        """Get all lines of a list of log files.

        :param log_id: The log id as defined in the configuration
        :param actions: An action list to search_on"""
        log_id = self.context_var("log_id")
        actions = [act.strip() for act in self.context_var("actions", []).split(",")]
        extras = [extra.strip() for extra in self.context_var("extras", []).split(",")]
        logs = get_logrotate_filenames(LOG_DIR, get_log_option(log_id, "audit_log"), suffix_regex=r"\.\d+$", full=True)
        for line in get_all_lines_of(logs, actions=actions):
            yield get_lines_info(line, extras)



from imio.fpaudit.browser.settings import IFPAuditSettings
from imio.fpaudit.interfaces import ILogsStorage
from imio.fpaudit.logger import FPAuditLogInfo
from imio.fpaudit.testing import IMIO_FPAUDIT_INTEGRATION_TESTING
from plone import api
from z3c.form import validator
from zope.component import getUtility
from zope.interface import Invalid

import unittest


class TestSettings(unittest.TestCase):

    layer = IMIO_FPAUDIT_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]

    def test_validate_settings(self):
        """Check invariant"""
        invariants = validator.InvariantsValidator(None, None, None, IFPAuditSettings, None)
        # test id uniqueness
        data = {
            "log_entries": [
                {"log_id": u"a", "audit_log": u"a.log", "log_format": u"%(asctime)s - %(message)s"},
                {"log_id": u"b", "audit_log": u"b.log", "log_format": u"%(asctime)s - %(message)s"},
            ]
        }
        self.assertFalse(invariants.validate(data))
        data = {
            "log_entries": [
                {"log_id": u"a", "audit_log": u"a.log", "log_format": u"%(asctime)s - %(message)s"},
                {"log_id": u"a", "audit_log": u"b.log", "log_format": u"%(asctime)s - %(message)s"},
            ]
        }
        errors = invariants.validate(data)
        self.assertTrue(isinstance(errors[0], Invalid))

    def test_settings_changed(self):
        """Check event"""
        api.portal.set_registry_record("imio.fpaudit.settings.log_entries", [])
        storage = getUtility(ILogsStorage)
        self.assertDictEqual(storage.storage, {})
        self.assertIsNone(storage.get("a"))
        api.portal.set_registry_record(
            "imio.fpaudit.settings.log_entries",
            [{"log_id": u"a", "audit_log": u"a.log", "log_format": u"%(asctime)s - %(message)s"}],
        )
        log_i = storage.get("a")
        self.assertIsNotNone(log_i)
        self.assertTrue(isinstance(log_i, FPAuditLogInfo))



# -*- coding: utf-8 -*-
from plone.app.testing import applyProfile
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from zope.globalrequest.local import setLocal

import imio.fpaudit
import os


class ImioFPAuditLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE,)

    def setUpZope(self, app, configurationContext):
        # Load any other ZCML that is required for your tests.
        # The z3c.autoinclude feature is disabled in the Plone fixture base layer.
        self.loadZCML(package=imio.fpaudit)

    def setUpPloneSite(self, portal):
        setLocal("request", portal.REQUEST)  # to avoid error with empty request in P6
        applyProfile(portal, "imio.fpaudit:default")
        setRoles(portal, TEST_USER_ID, ["Manager"])


IMIO_FPAUDIT_FIXTURE = ImioFPAuditLayer()


IMIO_FPAUDIT_INTEGRATION_TESTING = IntegrationTesting(
    bases=(IMIO_FPAUDIT_FIXTURE,), name="ImioFPAuditLayer:IntegrationTesting"
)


IMIO_FPAUDIT_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(IMIO_FPAUDIT_FIXTURE,), name="ImioFPAuditLayer:FunctionalTesting"
)


def write_temp_files(temp_dir, filenames):
    for filename in filenames:
        with open(os.path.join(temp_dir, filename), "w") as f:
            f.write("test")


def clear_temp_dir(temp_dir):
    for filename in os.listdir(temp_dir):
        file_path = os.path.join(temp_dir, filename)
        if os.path.isfile(file_path):
            os.remove(file_path)



# -*- coding: utf-8 -*-
from collective.z3cform.datagridfield.datagridfield import DataGridFieldFactory
from collective.z3cform.datagridfield.registry import DictRow
from imio.fpaudit import _
from imio.fpaudit.storage import store_config
from plone.app.registry.browser.controlpanel import ControlPanelFormWrapper
from plone.app.registry.browser.controlpanel import RegistryEditForm
from plone.autoform.directives import widget
from plone.registry.interfaces import IRecordModifiedEvent
from plone.z3cform import layout
from z3c.form.validator import NoInputData
from zope import schema
from zope.interface import Interface
from zope.interface import Invalid
from zope.interface import invariant


class ILogInfoSchema(Interface):
    log_id = schema.TextLine(
        title=_("Log id"),
        required=True,
    )

    audit_log = schema.TextLine(
        title=_("Audit file name"),
        description=_("Will be located in var/log"),
        required=True,
    )

    log_format = schema.TextLine(
        title=_("Log format"),
        default=u"%(asctime)s - %(message)s",
        required=True,
    )


class IFPAuditSettings(Interface):

    log_entries = schema.List(
        title=_(""),
        required=False,
        value_type=DictRow(title=_(u"Field"), schema=ILogInfoSchema, required=False),
    )
    widget(
        "log_entries",
        DataGridFieldFactory,
        display_table_css_class="listing",
        allow_reorder=False,
        auto_append=False,
    )

    @invariant
    def validate_settings(data):  # noqa
        # check ITableListSchema id uniqueness
        ids = []
        try:
            values = getattr(data, "log_entries") or []
        except NoInputData:
            return
        for entry in values:
            if entry["log_id"] in ids:
                raise Invalid(
                    _(
                        u"You cannot use same id multiple times '${id}'",
                        mapping={"id": entry["log_id"]},
                    )
                )
            ids.append(entry["log_id"])


class FPAuditSettings(RegistryEditForm):

    schema = IFPAuditSettings
    schema_prefix = "imio.fpaudit.settings"
    label = _("FP Audit Settings")


FPAuditSettingsView = layout.wrap_form(FPAuditSettings, ControlPanelFormWrapper)


def settings_changed(event):
    """Manage a record change"""
    if (
        IRecordModifiedEvent.providedBy(event)
        and event.record.interfaceName
        and event.record.interface != IFPAuditSettings
    ):
        return
    if event.record.fieldName == "log_entries":
        store_config(event.record.value)



# -*- coding: utf-8 -*-
from collective.fingerpointing.utils import get_request_information
from file_read_backwards import FileReadBackwards
from imio.fpaudit.interfaces import ILogsStorage
from natsort import natsorted
from plone import api
from zope.component import getUtility

import logging
import os
import re


logger = logging.getLogger("imio.fpaudit")
AUDIT_MESSAGE = u'user={0} ip={1} action={2} {3}'


def fplog(log_id, action, extras):
    """collective.fingerpointing add log message.

    :param log_id: The log id as defined in the configuration
    :param action: The action string
    :param extras: The extras"""
    user, ip = get_request_information()
    storage = getUtility(ILogsStorage)
    log_i = storage.get(log_id)
    if log_i:
        log_i(AUDIT_MESSAGE.format(user, ip, action, extras))
    else:
        logger.info(AUDIT_MESSAGE.format(user, ip, action, extras))


def get_all_lines_of(logfiles, actions=()):
    """Get all lines of a list of log files.

    :param logfiles: The list of log files
    :param actions: An action list to search_on"""
    for logfile in logfiles:
        for line in get_lines_of(logfile, actions=actions):
            yield line


def get_lines_info(line, extras):
    """Get the columns info of a log line.

    :param line: The log line
    :param extras: The extras tags list
    :return: A dictionary with the columns info"""
    dic = {}
    # 24-10-10 14:59:07 - user=admin ip=127.0.0.1 action=AUDIT col_a=xxxx col_b=yyy
    pattern = (
        r"(?P<date>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}) - user=(?P<user>.+?) "
        r"ip=(?P<ip>[\d\.]+|None) action=(?P<action>.+?)"
    )

    for extra in extras:
        pattern += r" {}=(?P<{}>.+?)".format(extra, extra)

    pattern += r"$"
    match = re.match(pattern, line)
    if match:
        dic = match.groupdict()
    else:
        dic["_line_"] = line
    return dic


def get_lines_of(logfile, actions=()):
    """Generator for reversed log lines of a log file.

    :param logfile: The path to the log file
    :param actions: An action list to search_on"""
    acts = [" action={}".format(act) for act in actions]
    with FileReadBackwards(logfile, encoding="utf-8") as file:
        for line in file:
            s_line = line.strip("\n")
            if not actions or any(act in s_line for act in acts):
                yield s_line


def get_log_option(log_id, option):
    """Get option value from settings

    :param log_id: The log id as defined in the configuration
    :param option: The option name"""
    log_entries = api.portal.get_registry_record("imio.fpaudit.settings.log_entries", default=[])
    for entry in log_entries:
        if entry["log_id"] == log_id:
            return entry.get(option)


def get_logrotate_filenames(directory, base_filename, suffix_regex=r"\.\d+$", full=True):
    """Get all logrotate files matching the base filename in the directory.

    :param directory: The directory where the logrotate files are stored
    :param base_filename: The base filename of the logrotate files
    :param suffix_regex: The regex pattern to match the suffix of the logrotate files
    :param full: If True, return the full path of the logrotate files
    """
    pattern = re.compile("^{}(?:{})*$".format(re.escape(base_filename), suffix_regex))
    res = natsorted([f for f in os.listdir(directory) if pattern.match(f)])
    if full:
        res = [os.path.join(directory, f) for f in res]
    return res
